@@ -20,9 +20,11 @@
  *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <unistd.h>
 #include <string.h>
 #include <stdio.h>
+#include <sys/param.h>
 #include <glib.h>
 #include <gnome.h>
 #include <libgnomevfs/gnome-vfs-mime.h>
@@ -31,8 +33,10 @@
 #include "file-utils.h"
 #include "fr-archive.h"
 #include "fr-command.h"
+#include "fr-command-ar.h"
 #include "fr-command-arj.h"
 #include "fr-command-cfile.h"
+#include "fr-command-iso.h"
 #include "fr-command-lha.h"
 #include "fr-command-rar.h"
 #include "fr-command-rpm.h"
@@ -40,19 +44,23 @@
 #include "fr-command-unstuff.h"
 #include "fr-command-zip.h"
 #include "fr-command-zoo.h"
+#include "fr-command-7z.h"
 #include "fr-error.h"
 #include "fr-marshal.h"
 #include "fr-process.h"
 #include "utf8-fnmatch.h"
 
 
-#define MAX_CHUNK_LEN 16000 /* FIXME : what is the max length of a command 
-			     * line ? */
+#define g_signal_handlers_disconnect_by_data(instance, data) \
+    g_signal_handlers_disconnect_matched ((instance), G_SIGNAL_MATCH_DATA, \
+                                          0, 0, NULL, NULL, (data))
+
+#define MAX_CHUNK_LEN (NCARGS - PATH_MAX) /* Max length of the command line */
 #define UNKNOWN_TYPE "application/octet-stream"
 #define SAME_FS (FALSE)
-#define NO_BACKUP_FILES (TRUE)
-#define NO_DOT_FILES (TRUE)
-#define IGNORE_CASE (TRUE)
+#define NO_BACKUP_FILES (FALSE)
+#define NO_DOT_FILES (FALSE)
+#define IGNORE_CASE (FALSE)
 
 
 enum {
@@ -188,6 +196,7 @@ fr_archive_init (FRArchive *archive)
 	archive->filename = NULL;
 	archive->command = NULL;
 	archive->is_compressed_file = FALSE;
+	archive->can_create_compressed_file = FALSE;
 
 	archive->fake_load_func = NULL;
 	archive->fake_load_data = NULL;
@@ -271,10 +280,20 @@ create_command_from_mime_type (FRArchive  *archive,
 	} else if (is_mime_type (mime_type, "application/x-rpm")) {
 		archive->command = fr_command_rpm_new (archive->process,
 						       filename);
+	} else if (is_mime_type (mime_type, "application/x-cd-image")) {
+		archive->command = fr_command_iso_new (archive->process,
+						       filename);
+	} else if (is_mime_type (mime_type, "application/x-deb") ||
+		   is_mime_type (mime_type, "application/x-ar")) {
+		archive->command = fr_command_ar_new (archive->process, 
+						      filename);
+	} else if (is_mime_type (mime_type, "application/x-7zip")) {
+		archive->command = fr_command_7z_new (archive->process,
+						      filename);
 	} else 
 		return FALSE;
 
-	return TRUE;
+	return (archive->command != NULL);
 }
 
 
@@ -363,76 +382,152 @@ create_command_from_filename (FRArchive  *archive,
 		archive->command = fr_command_tar_new (archive->process, 
 						       filename, 
 						       FR_COMPRESS_PROGRAM_GZIP);
-	} else if (file_extension_is (filename, ".tar.bz2")
-		   || file_extension_is (filename, ".tbz2")) {
+		return (archive->command != NULL);
+	} 
+	
+	if (file_extension_is (filename, ".tar.bz2")
+	    || file_extension_is (filename, ".tbz2")) {
 		archive->command = fr_command_tar_new (archive->process, 
 						       filename, 
 						       FR_COMPRESS_PROGRAM_BZIP2);
-	} else if (file_extension_is (filename, ".tar.bz")
-		   || file_extension_is (filename, ".tbz")) {
+		return (archive->command != NULL);
+	} 
+	
+	if (file_extension_is (filename, ".tar.bz")
+	    || file_extension_is (filename, ".tbz")) {
 		archive->command = fr_command_tar_new (archive->process, 
 						       filename, 
 						       FR_COMPRESS_PROGRAM_BZIP);
-	} else if (file_extension_is (filename, ".tar.Z")
-		   || file_extension_is (filename, ".taz")) {
+		return (archive->command != NULL);
+	} 
+	
+	if (file_extension_is (filename, ".tar.Z")
+	    || file_extension_is (filename, ".taz")) {
 		archive->command = fr_command_tar_new (archive->process, 
 						       filename, 
 						       FR_COMPRESS_PROGRAM_COMPRESS);
-	} else if (file_extension_is (filename, ".tar.lzo")
-		   || file_extension_is (filename, ".tzo")) {
+		return (archive->command != NULL);
+	}
+	
+	if (file_extension_is (filename, ".tar.lzo")
+	    || file_extension_is (filename, ".tzo")) {
 		archive->command = fr_command_tar_new (archive->process, 
 						       filename, 
 						       FR_COMPRESS_PROGRAM_LZOP);
-	} else if (file_extension_is (filename, ".tar")) {
+		return (archive->command != NULL);
+	} 
+	
+	if (file_extension_is (filename, ".tar")) {
 		archive->command = fr_command_tar_new (archive->process, 
 						       filename, 
 						       FR_COMPRESS_PROGRAM_NONE);
-	} else if (file_extension_is (filename, ".zip")
-		   || file_extension_is (filename, ".ear")
-		   || file_extension_is (filename, ".jar")
-		   || file_extension_is (filename, ".war")) {
+		return (archive->command != NULL);
+	} 
+
+	if (file_extension_is (filename, ".zip")
+	    || file_extension_is (filename, ".ear")
+	    || file_extension_is (filename, ".jar")
+	    || file_extension_is (filename, ".war")) {
 		archive->command = fr_command_zip_new (archive->process, 
 						       filename);
-	} else if (file_extension_is (filename, ".zoo")) {
+		return (archive->command != NULL);
+	} 
+	
+	if (file_extension_is (filename, ".zoo")) {
 		archive->command = fr_command_zoo_new (archive->process,
 						       filename);
-	} else if (file_extension_is (filename, ".lzh")) {
+		return (archive->command != NULL);
+	} 
+	
+	if (file_extension_is (filename, ".lzh")
+	    || file_extension_is (filename, ".lha")) {
 		archive->command = fr_command_lha_new (archive->process, 
 						       filename);
-	} else if (file_extension_is (filename, ".rar")) {
+		return (archive->command != NULL);
+	} 
+	
+	if (file_extension_is (filename, ".rar")) {
 		archive->command = fr_command_rar_new (archive->process, 
 						       filename);
-	} else if (file_extension_is (filename, ".arj")) {
+		return (archive->command != NULL);
+	} 
+
+	if (file_extension_is (filename, ".arj")) {
 		archive->command = fr_command_arj_new (archive->process, 
 						       filename);
-	} else if (loading) {
+		return (archive->command != NULL);
+	} 
+
+	if (file_extension_is (filename, ".ar")) {
+		archive->command = fr_command_ar_new (archive->process, 
+						      filename);
+		return (archive->command != NULL);
+	} 
+
+	if (file_extension_is (filename, ".7z")) {
+		archive->command = fr_command_7z_new (archive->process,
+						      filename);
+		return (archive->command != NULL);
+	} 
+
+	if (loading || archive->can_create_compressed_file) {
+
 		if (file_extension_is (filename, ".gz")
 		    || file_extension_is (filename, ".z")
 		    || file_extension_is (filename, ".Z")) {
 			archive->command = fr_command_cfile_new (archive->process, filename, FR_COMPRESS_PROGRAM_GZIP);
 			archive->is_compressed_file = TRUE;
-		} else if (file_extension_is (filename, ".bz")) {
+			return (archive->command != NULL);
+		} 
+
+		if (file_extension_is (filename, ".bz")) {
 			archive->command = fr_command_cfile_new (archive->process, filename, FR_COMPRESS_PROGRAM_BZIP);
 			archive->is_compressed_file = TRUE;
-		} else if (file_extension_is (filename, ".bz2")) {
+			return (archive->command != NULL);
+		} 
+
+		if (file_extension_is (filename, ".bz2")) {
 			archive->command = fr_command_cfile_new (archive->process, filename, FR_COMPRESS_PROGRAM_BZIP2);
 			archive->is_compressed_file = TRUE;
-		} else if (file_extension_is (filename, ".lzo")) {
+			return (archive->command != NULL);
+		}
+
+		if (file_extension_is (filename, ".lzo")) {
 			archive->command = fr_command_cfile_new (archive->process, filename, FR_COMPRESS_PROGRAM_LZOP);
 			archive->is_compressed_file = TRUE;
-		} else if (file_extension_is (filename, ".bin")
-			   || file_extension_is (filename, ".sit")) {
+			return (archive->command != NULL);
+		} 
+	} 
+
+	if (loading) {
+		
+		if (file_extension_is (filename, ".bin")
+		    || file_extension_is (filename, ".sit")) {
 			archive->command = fr_command_unstuff_new (archive->process,
 								   filename);
-		} else if (file_extension_is (filename, ".rpm")) {
+			return (archive->command != NULL);
+		} 
+
+		if (file_extension_is (filename, ".rpm")) {
 			archive->command = fr_command_rpm_new (archive->process,
 							       filename);
-		} else
-			return FALSE;
-	} else
-		return FALSE;
-	
-	return TRUE;
+			return (archive->command != NULL);
+		} 
+
+		if (file_extension_is (filename, ".iso")) {
+			archive->command = fr_command_iso_new (archive->process,
+							       filename);
+			return (archive->command != NULL);
+		} 
+		
+		if (file_extension_is (filename, ".deb")) {
+			archive->command = fr_command_ar_new (archive->process, 
+							      filename);
+			return (archive->command != NULL);
+		} 
+	}
+
+	return FALSE;
 }
 
 
@@ -477,7 +572,7 @@ action_performed (FRCommand   *command,
 		s_action = "Get list";
 		break;
 	}
-	g_print ("%s [DONE]\n", s_action);
+	g_print ("%s [DONE] (FR::Archive)\n", s_action);
 #endif
 
 	g_signal_emit (G_OBJECT (archive), 
@@ -625,8 +720,10 @@ fr_archive_load (FRArchive   *archive,
                         return FALSE;
 		}
 	
-	if (tmp_command != NULL) 
-		g_object_unref (G_OBJECT (tmp_command));
+	if (tmp_command != NULL) {
+		g_signal_handlers_disconnect_by_data (tmp_command, archive);
+		g_object_unref (tmp_command);
+	}
 
 	g_signal_connect (G_OBJECT (archive->command), 
 			  "start",
@@ -720,7 +817,7 @@ create_tmp_base_dir (const char *base_dir,
 	g_print ("dest_dir: %s\n", dest_dir);
 #endif
 
-	temp_dir = get_temp_work_dir_name ();
+	temp_dir = get_temp_work_dir ();
 	tmp = remove_level_from_path (dest_dir);
 	parent_dir =  g_build_filename (temp_dir, tmp, NULL);
 	g_free (tmp);
@@ -754,6 +851,8 @@ find_file_in_archive (FRArchive *archive,
 {
 	GList *scan;
 
+	g_return_val_if_fail (path != NULL, NULL);
+
 	for (scan = archive->command->file_list; scan; scan = scan->next) {
 		FileData *fdata = scan->data;
 		if (strcmp (path, fdata->original_path) == 0)
@@ -768,14 +867,33 @@ static void _archive_remove (FRArchive *archive, GList *file_list);
 
 
 static GList *
-escape_file_list (GList *file_list) 
+escape_file_list (FRCommand *command,
+		  GList     *file_list) 
 {
 	GList *e_file_list = NULL;
 	GList *scan;
 
 	for (scan = file_list; scan; scan = scan->next) {
 		char *filename = scan->data;
-		e_file_list = g_list_prepend (e_file_list, shell_escape (filename));
+		char *escape = fr_command_escape (command, filename);
+		e_file_list = g_list_prepend (e_file_list, escape);
+	}
+
+	return e_file_list;
+}
+
+
+static GList *
+shell_escape_file_list (FRCommand *command,
+			GList     *file_list) 
+{
+	GList *e_file_list = NULL;
+	GList *scan;
+
+	for (scan = file_list; scan; scan = scan->next) {
+		char *filename = scan->data;
+		char *escape = shell_escape (filename);
+		e_file_list = g_list_prepend (e_file_list, escape);
 	}
 
 	return e_file_list;
@@ -922,7 +1040,7 @@ fr_archive_add (FRArchive     *archive,
 
 	/* add now. */
 
-	e_file_list = escape_file_list (new_file_list);
+	e_file_list = shell_escape_file_list (archive->command, new_file_list);
 	fr_command_set_n_files (archive->command, g_list_length (e_file_list));
 
 	for (scan = e_file_list; scan != NULL; ) {
@@ -985,15 +1103,16 @@ file_list_remove_from_pattern (GList      **list,
 	for (scan = *list; scan;) {
 		char *path = scan->data;
 		char *utf8_name;
-
+		
 		utf8_name = g_filename_to_utf8 (file_name_from_path (path), 
 						-1, NULL, NULL, NULL);
-		if (match_patterns (patterns, utf8_name, FNM_CASEFOLD)) {
+
+		if (match_patterns (patterns, utf8_name, 0)) {
 			*list = g_list_remove_link (*list, scan);
 			g_free (scan->data);
 			g_list_free (scan);
 			scan = *list;
-		} else
+		} else 
 			scan = scan->next;
 
 		g_free (utf8_name);
@@ -1054,11 +1173,6 @@ fr_archive_add_with_wildcard (FRArchive     *archive,
 			      gboolean       update,
 			      gboolean       recursive,
 			      gboolean       follow_links,
-			      /*gboolean       same_fs,
-			      gboolean       no_backup_files,
-			      gboolean       no_dot_files,
-			      gboolean       ignore_case,
-			      */
 			      const char    *password,
 			      FRCompression  compression,
 			      DoneFunc       done_func,
@@ -1199,7 +1313,7 @@ _archive_remove (FRArchive *archive,
 		file_list_created = TRUE;
 	}
 
-	e_file_list = escape_file_list (file_list);
+	e_file_list = escape_file_list (archive->command, file_list);
 	fr_command_set_n_files (archive->command, g_list_length (e_file_list));
 
 	for (scan = e_file_list; scan != NULL; ) {
@@ -1264,21 +1378,23 @@ move_files_to_dir (FRArchive  *archive,
 	fr_process_add_arg (archive->process, "-f");
 	for (scan = file_list; scan; scan = scan->next) {
 		char  path[4096]; /* FIXME : 4096 ? */
-		char *filename = scan->data;
+		char *e_filename = shell_escape (scan->data);
 		
-		if (filename[0] == '/')
-			sprintf (path, "%s%s", source_dir, filename);
+		if (e_filename[0] == '/')
+			sprintf (path, "%s%s", source_dir, e_filename);
 		else
-			sprintf (path, "%s/%s", source_dir, filename);
+			sprintf (path, "%s/%s", source_dir, e_filename);
 		
 		fr_process_add_arg (archive->process, path);
+
+		g_free (e_filename);
 	}
 	fr_process_add_arg (archive->process, dest_dir);
 	fr_process_end_command (archive->process);
 }
 
 
-/* Note: all paths escaped, temp_dir and dest_dir unescaped. */
+/* Note: all paths unescaped, temp_dir and dest_dir unescaped. */
 static void
 move_files_in_chunks (FRArchive  *archive,
 		      GList      *file_list,
@@ -1317,7 +1433,7 @@ move_files_in_chunks (FRArchive  *archive,
 				   e_dest_dir);
 		prev->next = scan;
 	}	
-	
+
 	g_free (e_temp_dir);
 	g_free (e_dest_dir);
 }
@@ -1377,6 +1493,98 @@ extract_in_chunks (FRCommand  *command,
 }
 
 
+static char*
+compute_base_path (const char *base_dir,
+		   const char *path,
+		   gboolean    junk_paths,
+		   gboolean    can_junk_paths)
+{
+	int         base_dir_len = strlen (base_dir);
+	int         path_len = strlen (path);
+	const char *base_path;
+	char       *name_end;
+	char       *new_path;
+
+	if (junk_paths) {
+		if (can_junk_paths)
+			new_path = g_strdup (file_name_from_path (path));
+		else
+			new_path = g_strdup (path);
+#ifdef DEBUG
+		g_print ("%s, %s --> %s\n", base_dir, path, new_path);
+#endif
+		return new_path;
+	}
+
+	if (path_len <= base_dir_len)
+		return NULL;
+
+	base_path = path + base_dir_len;
+	if (path[0] != '/')
+		base_path -= 1;
+	name_end = strchr (base_path, '/');
+
+	if (name_end == NULL)
+		new_path = g_strdup (path);
+	else {
+		int name_len = name_end - path;
+		new_path = g_strndup (path, name_len);
+	}
+
+#ifdef DEBUG
+	g_print ("%s, %s --> %s\n", base_dir, path, new_path);
+#endif
+
+	return new_path;
+}
+
+
+static GList*
+compute_list_base_path (const char *base_dir,
+			GList      *filtered,
+			gboolean    junk_paths,
+			gboolean    can_junk_paths)
+{
+	GList *scan;
+	GList *list = NULL, *list_unique = NULL;
+	GList *last_inserted;
+
+	if (filtered == NULL)
+		return NULL;
+
+	for (scan = filtered; scan; scan = scan->next) {
+		const char *path = scan->data;
+		char       *new_path;
+		new_path = compute_base_path (base_dir, path, junk_paths, can_junk_paths);
+		if (new_path != NULL)
+			list = g_list_prepend (list, new_path);
+	}
+
+	/* The above operation can create duplicates, we remove them here. */
+	list = g_list_sort (list, (GCompareFunc)strcmp);
+	
+	last_inserted = NULL;
+	for (scan = list; scan; scan = scan->next) {
+		const char *path = scan->data;
+
+		if (last_inserted != NULL) {
+			const char *last_path = (const char*)last_inserted->data;
+			if (strcmp (last_path, path) == 0) {
+				g_free (scan->data);
+				continue;
+			}
+		}
+
+		last_inserted = scan;
+		list_unique = g_list_prepend (list_unique, scan->data);
+	}
+
+	g_list_free (list);
+
+	return list_unique;
+}
+
+
 /* Note : All paths unescaped.  
  * Note2: Do not escape dest_dir it will escaped in fr_command_extract if 
  *        needed. */
@@ -1384,6 +1592,7 @@ void
 fr_archive_extract (FRArchive  *archive,
 		    GList      *file_list,
 		    const char *dest_dir,
+		    const char *base_dir,
 		    gboolean    skip_older,
 		    gboolean    overwrite,
 		    gboolean    junk_paths,
@@ -1392,7 +1601,9 @@ fr_archive_extract (FRArchive  *archive,
 	GList    *filtered, *e_filtered;
 	GList    *scan;
 	gboolean  extract_all;
+	gboolean  use_base_dir;
 	gboolean  move_to_dest_dir;
+	gboolean  file_list_created = FALSE;
 
 	g_return_if_fail (archive != NULL);
 
@@ -1401,12 +1612,30 @@ fr_archive_extract (FRArchive  *archive,
 	/* if a command supports all the requested options use 
 	 * fr_command_extract directly. */
 
-	if (! (! overwrite && ! archive->command->propExtractCanAvoidOverwrite)
+	extract_all = (file_list == NULL);
+	if (extract_all && ! archive->command->propCanExtractAll) {
+		GList *scan;
+
+		scan = archive->command->file_list;
+		for (; scan; scan = scan->next) {
+			FileData *fdata = scan->data;
+			file_list = g_list_prepend (file_list, g_strdup (fdata->original_path));
+		}
+
+		file_list_created = TRUE;
+	}
+
+	use_base_dir = ! ((base_dir == NULL) 
+			  || (strcmp (base_dir, "") == 0)
+			  || (strcmp (base_dir, "/") == 0));
+	
+	if (! use_base_dir
+	    && ! (! overwrite && ! archive->command->propExtractCanAvoidOverwrite)
 	    && ! (skip_older && ! archive->command->propExtractCanSkipOlder)
 	    && ! (junk_paths && ! archive->command->propExtractCanJunkPaths)) {
 		GList *e_file_list;
 
-		e_file_list = escape_file_list (file_list);
+		e_file_list = escape_file_list (archive->command, file_list);
 		extract_in_chunks (archive->command,
 				   e_file_list,
 				   dest_dir,
@@ -1416,16 +1645,19 @@ fr_archive_extract (FRArchive  *archive,
 				   password);
 		path_list_free (e_file_list);
 
+		if (file_list_created) 
+			path_list_free (file_list);
+
 		return;
 	}
 
 	/* .. else we have to implement the unsupported options. */
+	
+	move_to_dest_dir = (use_base_dir
+			    || ((junk_paths 
+				 && ! archive->command->propExtractCanJunkPaths)));
 
-	move_to_dest_dir = (junk_paths 
-			    && ! archive->command->propExtractCanJunkPaths);
-
-	extract_all = (file_list == NULL);
-	if (extract_all) {
+	if (extract_all && ! file_list_created) {
 		GList *scan;
 
 		scan = archive->command->file_list;
@@ -1433,6 +1665,8 @@ fr_archive_extract (FRArchive  *archive,
 			FileData *fdata = scan->data;
 			file_list = g_list_prepend (file_list, g_strdup (fdata->original_path));
 		}
+
+		file_list_created = TRUE;
 	}
 
 	filtered = NULL;
@@ -1493,14 +1727,13 @@ fr_archive_extract (FRArchive  *archive,
 		return;
 	} 
 
-	e_filtered = escape_file_list (filtered);	
+	e_filtered = escape_file_list (archive->command, filtered);	
 
 	if (move_to_dest_dir) {
 		char *temp_dir;
 		char *e_temp_dir;
 
-		temp_dir = get_temp_work_dir_name ();
-		ensure_dir_exists (temp_dir, 0700);
+		temp_dir = get_temp_work_dir ();
 		extract_in_chunks (archive->command,
 				   e_filtered,
 				   temp_dir,
@@ -1508,8 +1741,15 @@ fr_archive_extract (FRArchive  *archive,
 				   skip_older,
 				   junk_paths,
 				   password);
+
+		if (use_base_dir) {
+			GList *tmp_list = compute_list_base_path (base_dir, filtered, junk_paths, archive->command->propExtractCanJunkPaths);
+			g_list_free (filtered);
+			filtered = tmp_list;
+		}
+
 		move_files_in_chunks (archive, 
-				      e_filtered, 
+				      filtered, 
 				      temp_dir, 
 				      dest_dir);
 
@@ -1536,7 +1776,7 @@ fr_archive_extract (FRArchive  *archive,
 	if (filtered != NULL)
 		g_list_free (filtered);
 
-	if (extract_all) 
+	if (file_list_created) 
 		/* the list has been created in this function. */
 		path_list_free (file_list);
 }
@@ -1564,12 +1804,16 @@ G_CONST_RETURN char *
 fr_archive_utils__get_file_name_ext (const char *filename)
 {
 	static char * ext[] = {
+		".7z",
+		".ar",
 		".arj",
 		".bin",
 		".bz", 
 		".bz2", 
+		".deb",
 		".ear",
 		".gz", 
+		".iso",
 		".jar",
 		".lzh",
 		".lzo",
