@@ -590,7 +590,10 @@ _gtk_entry_set_filename_text (GtkEntry   *entry,
 	char *utf8_text;
 
 	utf8_text = g_filename_to_utf8 (text, -1, NULL, NULL, NULL);
-	gtk_entry_set_text (entry, utf8_text);
+	if (utf8_text != NULL)
+		gtk_entry_set_text (entry, utf8_text);
+	else
+		gtk_entry_set_text (entry, "");
 	g_free (utf8_text);
 }
 
@@ -617,7 +620,7 @@ _gtk_label_set_filename_text (GtkLabel   *label,
 {
 	char *utf8_text;
 
-	utf8_text = g_filename_to_utf8 (text, -1, NULL, NULL, NULL);
+	utf8_text = g_filename_display_name (text);
 	gtk_label_set_text (label, utf8_text);
 	g_free (utf8_text);
 }
